import os
import sys

from pyramid.paster import get_appsettings, setup_logging
from pyramid.scripts.common import parse_vars

from ..models.meta import Base
from ..models import get_engine


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri> [var=value]\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    if len(argv) < 2:
        usage(argv)
    config_uri = argv[1]
    options = parse_vars(argv[2:])
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, options=options)
    if 'DATABASE_URL' in os.environ:
        settings['sqlalchemy.url'] = os.environ['DATABASE_URL']

    engine = get_engine(settings)
    Base.metadata.create_all(engine)



from os import environ

from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.security import Allow, Authenticated, Everyone
from pyramid.session import SignedCookieSessionFactory


class PyramidStocksRoot:
    def __init__(self, request):
        """
        Initialize for access control.
        """
        self.request = request

    __acl__ = [
        (Allow, Everyone, 'view'),
        (Allow, Authenticated, 'secret'),
    ]


def includeme(config):
    auth_secret = environ.get(
        'AUTH_SECRET',
        '\x85v\xaf\xf9:3I\x0c\x98\x80\xec\x9e\xe3\xd4\xee\xba\xdfst\xb5')
    authz_policy = ACLAuthorizationPolicy()
    authn_policy = AuthTktAuthenticationPolicy(
        secret=auth_secret,
        hashalg='sha512',
    )

    config.set_authentication_policy(authn_policy)
    config.set_authorization_policy(authz_policy)
    config.set_default_permission('view')
    config.set_root_factory(PyramidStocksRoot)

    session_secret = environ.get(
        'SESSION_SECRET',
        '\xc0\xa37\xc6\x15\xb4B7\x16\xe9\xaa[^<`\x7f;p\x18\x1f')
    session_factory = SignedCookieSessionFactory(session_secret)

    config.set_session_factory(session_factory)
    config.set_default_csrf_options(require_csrf=True)



def test_default_behavior_of_home_view(dummy_request):
    from ..views.home import home_view

    response = home_view(dummy_request)
    assert dummy_request.response.status_code == 200
    assert isinstance(response, dict)
    assert response == {}


def test_not_found_behavior(dummy_request):
    from ..views.notfound import notfound_view

    response = notfound_view(dummy_request)
    assert dummy_request.response.status_code == 404
    assert isinstance(response, dict)
    assert response == {}



from concurrent.futures import ProcessPoolExecutor
from functools import partial
from itertools import groupby, repeat
from operator import itemgetter, methodcaller
from random import sample
from itertools import chain

from .base_agent import BaseAgent


def call(_call, *args, **kwargs):
    return _call(*args, **kwargs)


class Agent(BaseAgent):
    '''Computer Agent'''

    def get_boards(self, cursor):
        '''Retrieves potential board states'''
        params = {'lookahead': self.lookahead}
        if cursor:
            params['cursor'] = cursor
        return self.request('GET', '/v1.0/games/{}/states'.format(self.game_id), params=params)

    def get_boards_cursor(self):
        cursor = True
        while cursor:
            board_options = self.get_boards(cursor)
            cursor = board_options['cursor']
            yield tuple(map(
                lambda boards: tuple(map(
                    lambda board: tuple(map(bytes.fromhex, board)),
                    boards)),
                board_options['boards']))

    def play_round(self):
        '''Play a game round'''
        with ProcessPoolExecutor(4) as executor:
            # best_boards = [(root_value, root), ...]
            best_boards = executor.map(
                call,
                map(partial(methodcaller, 'evaluate_boards'), self.get_boards_cursor()),
                repeat(self),
                chunksize=50)
            # best_boards = [(root_value, [(root_value, root), ...]), ...]
            best_boards = groupby(sorted(best_boards), itemgetter(0))
            # _, best_boards = (root_value, [(root_value, root), ...])
            try:
                score, best_boards = next(best_boards)
                print(score)
            except StopIteration:
                return self.close()
            # best_boards = [root, ...]
            best_boards = list(map(
                next,
                map(
                    itemgetter(1),
                    groupby(chain.from_iterable(map(itemgetter(1), best_boards))))))
            if not best_boards:
                return self.close()
            return self.put_board(sample(best_boards, 1)[0])

    def play_game(self):
        '''Play a game'''
        game_over = False
        while not game_over:
            game_over = self.play_round()
        return game_over



from .meta import Base
from sqlalchemy.orm import relationship
from .table_ass import table
from sqlalchemy import (
    Column,
    String,
    Integer,
)


class TableBoard(Base):
    __tablename__ = 'board'
    id = Column(Integer, primary_key=True)
    board_state = Column(String, nullable=False)
    move_num = Column(Integer, nullable=False)
    player = Column(String, nullable=False)
    game = Column(String, nullable=False)
    game_link = relationship("TableGame", secondary=table, back_populates="board_link")



import os
from pyramid.config import Configurator

testapp = None


def main(global_config, **settings):
    """
    Return a Pyramid WSGI application.
    """
    if os.environ.get('DATABASE_URL', ''):
        settings['sqlalchemy.url'] = os.environ['DATABASE_URL']
    else:
        settings['sqlalchemy.url'] = 'postgres://localhost:5432/neuralknight'
    if os.environ.get('PORT', ''):
        settings['listen'] = '*:' + os.environ['PORT']
    else:
        settings['listen'] = 'localhost:8080'
    config = Configurator(settings=settings)
    config.include('cornice')
    config.include('pyramid_jinja2')
    config.include('.models')
    config.include('.routes')
    # config.include('.security')
    config.scan()
    return config.make_wsgi_app()



from ..models import BaseBoard, BaseAgent


class MockBoard(BaseBoard):
    def __init__(self, testapp, cursor=None):
        self.testapp = testapp
        self.args = {}
        self.kwargs = {}
        self.cursor = cursor
        super().__init__([[0 for i in range(8)] for j in range(8)])

    def slice_cursor_v1(self, *args, **kwargs):
        self.args['slice_cursor_v1'] = args
        self.kwargs['slice_cursor_v1'] = kwargs
        return {
            'cursor': self.cursor,
            'boards': [(self.board,)]
        }

    def add_player_v1(self, *args, **kwargs):
        self.args['add_player_v1'] = args
        self.kwargs['add_player_v1'] = kwargs
        player = args[1]
        if self.player1:
            self.player2 = player
        else:
            self.player1 = player
        self.poke_player(False)
        return {}

    def update_state_v1(self, *args, **kwargs):
        self.args['update_state_v1'] = args
        self.kwargs['update_state_v1'] = kwargs
        return {'end': True}


# def test_player_connection(testapp):
#     '''Assert players connect to board'''
#     mockboard = MockBoard(testapp)
#     player1 = testapp.post_json('/issue-agent', {'id': mockboard.id}).json
#     player2 = testapp.post_json('/issue-agent', {'id': mockboard.id, 'player': 2}).json
#     assert player1
#     assert player2


# this needs to change - need to check multi-gets
# def test_get_boards(testapp):
#     mockboard = MockBoard(testapp, 1)
#     player1 = testapp.post_json('/issue-agent', {'id': mockboard.id}).json
#     assert player1['agent_id'] in BaseAgent.AGENT_POOL
#     player2 = testapp.post_json('/issue-agent', {'id': mockboard.id, 'player': 2}).json
#     assert player2
#     assert player1['agent_id'] not in BaseAgent.AGENT_POOL


# def test_choose_valid_move(testapp):
#     '''Assert agent chooses valid move and game ends'''
#     mockboard = MockBoard(testapp)
#     state = mockboard.current_state_v1()
#     player1 = testapp.post_json('/issue-agent', {'id': mockboard.id}).json
#     # assert player1['agent_id'] in BaseAgent.AGENT_POOL
#     player2 = testapp.post_json('/issue-agent', {'id': mockboard.id, 'player': 2}).json
#     assert state == mockboard.current_state_v1()
#     assert player2
#     assert player1['agent_id'] not in BaseAgent.AGENT_POOL


# def test_play_game(testapp):
#     mockboard = MockBoard(testapp)
#     player1 = testapp.post_json('/issue-agent', {'id': mockboard.id}).json
#     # assert player1['agent_id'] in BaseAgent.AGENT_POOL
#     player2 = testapp.post_json('/issue-agent', {'id': mockboard.id, 'player': 2}).json
#     assert player2
#     assert player1['agent_id'] not in BaseAgent.AGENT_POOL


def test_user_connection(testapp):
    mockboard = MockBoard(testapp)
    player1 = testapp.post_json('/issue-agent', {'id': mockboard.id, 'user': True}).json
    assert player1



def includeme(config):
    config.add_static_view('static', 'static', cache_max_age=3600)
    config.add_route('home', '/')
    config.add_route('issue_agent', '/issue-agent')
    config.add_route('issue_agent_lookahead', '/issue-agent-lookahead')
    config.add_route('agent', '/agent/{agent_id}')



from pyramid.httpexceptions import HTTPBadRequest
from pyramid.view import view_config

from ..models import Agent
from ..models import UserAgent


@view_config(route_name='issue_agent', request_method='POST', renderer='json')
def issue_agent_view(request):
    try:
        game_id = request.json['id']
    except KeyError:
        return HTTPBadRequest()
    player = request.json.get('player', 1)
    if 'user' in request.json:
        return {'agent_id': UserAgent(game_id, player).agent_id}
    return {'agent_id': Agent(game_id, player).agent_id}


@view_config(route_name='issue_agent_lookahead', request_method='POST', renderer='json')
def issue_agent_lookahead_view(request):
    try:
        game_id = request.json['id']
        lookahead = request.json['lookahead']
    except KeyError:
        return HTTPBadRequest()
    player = request.json.get('player', 1)
    if 'user' in request.json:
        return {'agent_id': UserAgent(game_id, player).agent_id}
    return {'agent_id': Agent(game_id, player, lookahead).agent_id}


@view_config(route_name='agent', request_method=('PUT', 'GET'), renderer='json')
def agent_view(request):
    agent_id = request.matchdict['agent_id']
    try:
        agent = Agent.get_agent(agent_id)
    except KeyError:
        if request.method == 'GET':
            return {'state': {'end': True}}
        return {}

    if request.method == 'GET':
        return {'state': agent.get_state()}
    if request.json.get('end', False):
        return agent.close()
    if isinstance(agent, UserAgent):
        return agent.play_round(request.json.get('move', None))
    return agent.play_round()



"""
Chess state handling model.
"""

from itertools import chain, count, groupby, islice, starmap
from functools import partial, lru_cache
from operator import itemgetter
from uuid import uuid4

from .board_constants import (
    INITIAL_BOARD, BISHOP, KING, KNIGHT, QUEEN, ROOK, unit,
    BISHOP_MOVES, KING_MOVES, KNIGHT_MOVES, QUEEN_MOVES, ROOK_MOVES)

__all__ = ['BoardModel', 'CursorDelegate', 'InvalidMove']


class InvalidMove(RuntimeError):
    pass


class CursorDelegate:
    def __init__(self):
        self.cursors = {}

    def get_cursor(self, board, cursor, lookahead, complete):
        """
        Retrieve iterable for cursor.
        """
        if cursor in self.cursors:
            return self.cursors.pop(cursor)
        if complete:
            it = board.lookahead_boards(lookahead)
        else:
            it = board.prune_lookahead_boards(lookahead)
        it = groupby(it, itemgetter(0))
        return next(it)[1], it

    def slice_cursor_v1(self, board, cursor, lookahead, complete):
        """
        Retrieve REST cursor slice.
        """
        it, cur = self.get_cursor(board, cursor, lookahead, complete)
        slen = (900 // lookahead) if complete else 450
        boards = tuple(islice(it, slen))
        if len(boards) < slen:
            try:
                it = next(cur)[1]
            except StopIteration:
                return {'cursor': None, 'boards': boards}
        cursor = str(uuid4())
        self.cursors[cursor] = (it, cur)
        return {'cursor': cursor, 'boards': boards}


@lru_cache()
def is_on_board(posX, posY, move):
    """
    Validate a move against board bounds.
    """
    return 0 <= (posX + move[0]) < 8 and 0 <= (posY + move[1]) < 8


@lru_cache()
def validate_ending(board, posX, posY, move):
    """
    Validate a move against ending location.
    """
    return not active_piece(
        board[posY + move[1]][posX + move[0]])


@lru_cache()
def validate_move(board, posX, posY, move):
    """
    Validate clear path along move.
    """
    return (
        validate_ending(board, posX, posY, move)
        and all(
            map(
                lambda _range:
                    not board
                    [posY + unit(move[1]) * _range]
                    [posX + unit(move[0]) * _range],
                range(1, max(abs(move[0]), abs(move[1]))))))


@lru_cache()
def validation_for_piece(board, piece, posX, posY):
    """
    Get final validation function for piece.
    """
    def validate_true(*args):
        return True

    return partial((
        validate_true,  # No piece
        partial(validate_move, board),  # Bishop
        partial(validate_ending, board),  # King
        partial(validate_ending, board),  # Knight
        validate_true,  # Pawn
        partial(validate_move, board),  # Queen
        partial(validate_move, board)  # Rook
        )[(piece & 0xE) // 2], posX, posY)


def moves_for_pawn(board, piece, posX, posY):
    """
    Get all possible moves for pawn.
    """
    if (
            is_on_board(posX, posY, (0, -1))
            and (not board[posY - 1][posX])):
        yield (0, -1)
    if (
            posY == 6
            and (not board[posY - 1][posX])
            and (not board[posY - 2][posX])):
        yield (0, -2)
    if (
            is_on_board(posX, posY, (-1, -1))
            and inactive_piece(board[posY - 1][posX - 1])):
        yield (-1, -1)
    if (
            is_on_board(posX, posY, (1, -1))
            and inactive_piece(board[posY - 1][posX + 1])):
        yield (1, -1)
    if piece & 0x10:
        yield ()  # en passant


def moves_for_king(board, piece, posX, posY):
    """
    Get castling.
    """
    return
    if piece & 0x10:
        if (board[posY][0] & 0x10) and (board[posY][0] & 1) and (board[posY][0] & 0xE) == ROOK:
            yield (-2, 0)
            yield (-3, 0)
        if (board[posY][7] & 0x10) and (board[posY][7] & 1) and (board[posY][7] & 0xE) == ROOK:
            yield (2, 0)


@lru_cache()
def moves_for_piece(board, piece, posX, posY):
    """
    Get all possible moves for piece type.
    """
    return tuple(filter(partial(is_on_board, posX, posY), (
        (),  # No piece
        BISHOP_MOVES,
        chain(KING_MOVES, moves_for_king(board, piece, posX, posY)),
        KNIGHT_MOVES,
        moves_for_pawn(board, piece, posX, posY),
        QUEEN_MOVES,
        ROOK_MOVES
      )[(piece & 0xE) // 2]))


@lru_cache()
def valid_moves_for_piece(board, piece, posX, posY):
    """
    Get all valid moves for piece type.
    """
    return tuple(filter(
        validation_for_piece(board, piece, posX, posY),
        moves_for_piece(board, piece, posX, posY)))


@lru_cache()
def lookahead_boards_for_piece(board, check, piece, posX, posY):
    """
    Get all future board states.
    """
    def mutate_board(move):
        new_state = list(map(list, board))
        new_state[posY][posX] = 0
        if piece == 9 and posY == 1:
            for promote in (BISHOP, KNIGHT, QUEEN, ROOK):
                new_state[posY + move[1]][posX + move[0]] = promote | 1
                yield swap(tuple(map(bytes, new_state)))
        else:
            new_state[posY + move[1]][posX + move[0]] = piece & 0xF
            yield swap(tuple(map(bytes, new_state)))

    _valid_moves_for_piece = valid_moves_for_piece(board, piece, posX, posY)
    if check:
        _valid_moves_for_piece = filter(
            lambda move: board[posY + move[1]][posX + move[0]] & 0xE == KING,
            _valid_moves_for_piece)
    return tuple(chain.from_iterable(map(mutate_board, _valid_moves_for_piece)))


def lookahead_check_for_piece(board, piece, posX, posY):
    """
    Get possiblity of check in all future board states.
    """
    return map(
        lambda move: (board[posY + move[1]][posX + move[0]] & 0xF) == KING,
        valid_moves_for_piece(board, piece, posX, posY))


@lru_cache()
def lookahead_check(board):
    return any(chain.from_iterable(
        starmap(partial(lookahead_check_for_piece, board), active_pieces(board))))


@lru_cache()
def active_piece(piece):
    """
    Validate piece as active.
    """
    return piece & 1 and piece & 0xE


@lru_cache()
def inactive_piece(piece):
    """
    Validate piece as inactive.
    """
    return (not piece & 1) and piece & 0xE


@lru_cache()
def active_pieces(board):
    """
    Get all pieces for current player.
    """
    return tuple(chain.from_iterable(
        map(
            lambda posY, row: filter(None, map(
                lambda posX, piece:
                    (piece, posX, posY)
                    if active_piece(piece) else
                    None,
                count(), row)),
            count(), board)))


@lru_cache()
def swap(board):
    """
    Rotate active player.
    """
    return tuple(list(map(
        lambda row: bytes(map(
            lambda pp:
                (pp ^ 1) if pp else 0,
            row))[::-1],
        board))[::-1])


def make_tuple(*args):
    return args


class BoardModel:
    """
    Chess board model.
    """

    def __init__(self, board=None):
        """
        Set up board.
        """
        self.board = board or INITIAL_BOARD
        self.move_count = 1
        self.moves_since_pawn = 0

    def __bool__(self):
        """
        Ensure active player king on board.
        """
        return self.moves_since_pawn >= 50 or self.has_kings()

    def __contains__(self, piece):
        """
        Ensure piece on board.
        """
        return any(map(lambda row: piece in row or (piece | 0x10) in row, self.board))

    def swap(self):
        """
        Rotate active player.
        """
        return swap(self.board)

    def validate_mutation(self, mutation, state):
        if len(mutation) != 2:
            raise InvalidMove
        if mutation[0][3] == 0:
            old, new = mutation
        elif mutation[1][3] == 0:
            new, old = mutation
        else:
            raise InvalidMove
        new_posX, new_posY, new_prev_piece, new_next_piece = new
        new_next_piece = new_next_piece & 0xF
        if active_piece(new_prev_piece):
            raise InvalidMove
        posX, posY, moved_piece, _ = old
        piece = moved_piece & 0xF
        if not active_piece(piece):
            raise InvalidMove
        if piece != new_next_piece:
            if not (piece == 9 and new_posY == 0):
                raise InvalidMove
        move = (new_posX - posX, new_posY - posY)
        if move not in valid_moves_for_piece(self.board, moved_piece, posX, posY):
            raise InvalidMove
        if piece == 9:
            self.moves_since_pawn = 0
            if new_posY == 0 and new_next_piece == 9:
                state = list(map(list, state))
                state[new_posY][new_posX] = QUEEN | 1
                return swap(tuple(map(bytes, state)))
        return swap(state)

    def update(self, state):
        """
        Validate and return new board state.
        """
        mutation = tuple(filter(None, chain.from_iterable(map(
            lambda posY, old_row, new_row: map(
                lambda posX, old_piece, new_piece:
                    None
                    if old_piece == new_piece else
                    (posX, posY, old_piece, new_piece),
                count(),
                old_row, new_row),
            count(),
            self.board, state))))
        board = BoardModel(self.validate_mutation(mutation, state))
        board.move_count = self.move_count + 1
        board.moves_since_pawn = self.moves_since_pawn + 1
        return board

    def lookahead_boards_for_board(self, check):
        return chain.from_iterable(
            starmap(
                partial(lookahead_boards_for_piece, self.board, check),
                active_pieces(self.board)))

    def _valid_root_lookahead_boards_end(self):
        """
        Provide an iterable of valid moves for current board state.
        """
        check = lookahead_check(self.board)
        return all(map(
            lambda board: (KING | 1) in BoardModel(board),
            self.lookahead_boards_for_board(check)))

    def _valid_root_lookahead_boards(self, check):
        """
        Provide an iterable of valid moves for current board state.
        """
        return filter(
            lambda board: BoardModel(board)._valid_root_lookahead_boards_end(),
            self.lookahead_boards_for_board(check))

    def _prune_lookahead_boards(self, n, active=True):
        """
        Provide an iterable of valid moves for current board state.
        """
        check = lookahead_check(self.board)
        if n == 1:
            return iter((self.board if active else swap(self.board),))
        return chain.from_iterable(map(
            lambda board: BoardModel(board)._prune_lookahead_boards(n - 1, not active),
            self.lookahead_boards_for_board(check)))

    def prune_lookahead_boards(self, n):
        """
        Provide an iterable of valid moves for current board state.
        """
        check = lookahead_check(self.board)
        return chain.from_iterable(
            map(
                lambda board: map(
                    partial(make_tuple, swap(board)),
                    BoardModel(board)._prune_lookahead_boards(n - 1)),
                self._valid_root_lookahead_boards(check)))

    def _lookahead_boards(self, n, active=False):
        """
        Provide an iterable of valid moves for current board state.
        """
        check = lookahead_check(self.board)
        if not self:
            return iter((((self.board if active else swap(self.board)) for _ in range(n)),))
        if n == 0:
            return iter(((),))
        return chain.from_iterable(
            map(
                lambda board: map(
                    lambda t: (board if active else swap(board),) + tuple(t),
                    BoardModel(board)._lookahead_boards(n - 1, not active)),
                self.lookahead_boards_for_board(check)))

    def lookahead_boards(self, n, active=False):
        """
        Provide an iterable of valid moves for current board state.
        """
        check = lookahead_check(self.board)
        if not self:
            return iter((((self.board if active else swap(self.board)) for _ in range(n)),))
        if n == 0:
            return iter(((),))
        return chain.from_iterable(
            map(
                lambda board: map(
                    lambda t: (board if active else swap(board),) + tuple(t),
                    BoardModel(board)._lookahead_boards(n - 1, not active)),
                self._valid_root_lookahead_boards(check)))

    def has_kings(self):
        """
        Ensure active kings on board.
        """
        return (KING | 1) in self and KING in self



from pyramid.httpexceptions import HTTPBadRequest

from ..views.agent import issue_agent_view


def test_issue_agent_bad_request(dummy_post_request):
    '''Test request without proper query string'''
    assert isinstance(issue_agent_view(dummy_post_request), HTTPBadRequest)



from os import environ
from pyramid.testing import DummyRequest, setUp, tearDown
from pytest import fixture

from ..models.board import Board
from ..models.meta import Base


@fixture
def configuration(request):
    """
    Create database models for testing purposes.
    """
    config = setUp(settings={
        'sqlalchemy.url': environ.get(
            'TEST_DATABASE_URL', 'postgres://localhost:5432/testing_neuralknight')
    })
    config.include('neuralknight.models')
    config.include('neuralknight.routes')
    yield config
    tearDown()


@fixture
def db_session(configuration, request):
    """
    Create a database session for interacting with the test database.
    """
    SessionFactory = configuration.registry['dbsession_factory']
    session = SessionFactory()
    engine = session.bind
    Base.metadata.create_all(engine)
    yield session
    session.transaction.rollback()
    Base.metadata.drop_all(engine)


@fixture
def dummy_request(db_session):
    """
    Create a dummy GET request with a dbsession.
    """
    return DummyRequest(dbsession=db_session)


@fixture
def dummy_post_request(db_session):
    """
    Create a dummy POST request with a dbsession.
    """
    return DummyRequest(dbsession=db_session, post={}, json={})


@fixture(scope="session")
def testapp(request):
    """
    Functional test for app to support mocking.
    """
    import neuralknight
    from webtest import TestApp

    app = neuralknight.main({}, **{
        'sqlalchemy.url': environ.get(
            'TEST_DATABASE_URL', 'postgres://localhost:5432/testing_neuralknight')
    })

    SessionFactory = app.registry["dbsession_factory"]
    engine = SessionFactory().bind
    Base.metadata.create_all(bind=engine)
    neuralknight.testapp = TestApp(app)
    yield neuralknight.testapp
    Base.metadata.drop_all(bind=engine)


@fixture
def start_board():
    return Board()


@fixture
def pawn_capture_board():
    return Board(tuple(map(bytes, [
        [12, 6, 2, 10, 4, 2, 6, 12],
        [8, 8, 8, 0, 8, 0, 8, 8],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 8, 0, 8, 0, 0],
        [0, 0, 0, 0, 9, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [9, 9, 9, 9, 0, 9, 9, 9],
        [13, 7, 3, 11, 5, 3, 7, 13]])))


@fixture
def min_pawn_board():
    return Board(tuple(map(bytes, [
        [0, 0, 0, 0, 4, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 9, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 5, 0, 0, 0]])))


@fixture
def end_game_board():
    return Board(tuple(map(bytes, [
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 4, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 11, 0],
        [0, 0, 0, 5, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0]])))



from .base_agent import BaseAgent


class UserAgent(BaseAgent):
    '''Human Agent'''

    def __init__(self, game_id, player):
        super().__init__(game_id, player)

    def get_state(self):
        '''Gets current board state'''
        if self.game_over:
            return {'end': True}
        data = self.request('GET', f'/v1.0/games/{ self.game_id }')
        return data['state']

    def play_round(self, move):
        if move is None:
            return
        proposal = self.get_state()
        if isinstance(proposal, dict):
            return
        proposal = list(map(list, map(bytes.fromhex, proposal)))
        proposal[move[1][0]][move[1][1]] = proposal[move[0][0]][move[0][1]]
        proposal[move[0][0]][move[0][1]] = 0
        return self.put_board(tuple(map(bytes, proposal)))



"""
Chess state handling model.
"""

from concurrent.futures import ThreadPoolExecutor
from itertools import count
from json import dumps

from .base_board import BaseBoard, NoBoard
from .table_board import TableBoard
from .table_game import TableGame

__all__ = ['Board', 'NoBoard']


class Board(BaseBoard):
    """
    Chess board interaction model.
    """

    EMOJI = [
      '⌛', '‼',
      '♝', '♗', '♚', '♔', '♞', '♘', '♟', '♙', '♛', '♕', '♜', '♖', '▪', '▫']

    def __init__(self, board=None, _id=None, active_player=True):
        """
        Set up board.
        """
        super().__init__(board, _id, active_player)
        self.executor = ThreadPoolExecutor()

    def __repr__(self):
        """
        Output the raw view of board.
        """
        return f'Board({ self.board !r})'

    def __str__(self):
        """
        Output the emoji view of board.
        """
        if self._active_player:
            def piece_to_index(piece):
                return (piece & 0xF)
        else:
            def piece_to_index(piece):
                return (piece & 0xE) | (0 if piece & 1 else 1)

        return '\n'.join(map(
            lambda posY, row: ''.join(map(
                lambda posX, piece: self.EMOJI[
                    piece_to_index(piece)
                    if piece else
                    14 + ((posY + posX) % 2)],
                count(), row)),
            count(),
            self.board if self._active_player else reversed(
                [reversed(row) for row in self.board])))

    def add_player_v1(self, dbsession, player):
        """
        Player 2 joins game.
        """
        assert player
        if self.player1:
            self.player2 = player
            table_game = TableGame(
                game=self.id,
                player_one=self.player1,
                player_two=self.player2,
                one_won=True,
                two_won=True)
            table_board = TableBoard(
                board_state=dumps(tuple(map(tuple, self.board))),
                move_num=self._board.move_count,
                player=self.active_player(),
                game=self.id)
            table_board.game_link.append(table_game)
            dbsession.add(table_game)
            dbsession.add(table_board)
            self.poke_player(False)
            return {}
        self.player1 = player
        return {}

    def slice_cursor_v1(self, cursor=None, lookahead=1, complete=False):
        """
        Retrieve REST cursor slice.
        """
        return self.cursor_delegate.slice_cursor_v1(self._board, cursor, int(lookahead), complete)

    def update_state_v1(self, dbsession, state):
        """
        Make a move to a new state on the board.
        """
        moving_player = self.active_player()
        board = self.update(state)
        table_game = dbsession.query(TableGame).filter(
            TableGame.game == board.id).first()
        table_board = TableBoard(
            board_state=dumps(tuple(map(tuple, board.board))),
            move_num=board._board.move_count,
            player=board.active_player(),
            game=board.id)
        if table_game:  # TODO(grandquista)
            table_board.game_link.append(table_game)
        dbsession.add(table_board)
        if board:
            board.poke_player(False)
            return {'end': False}
        board.poke_player(True, moving_player)
        if board._board.has_kings():
            table_game.one_won = False
            table_game.two_won = False
        elif moving_player == table_game.player_one:
            table_game.two_won = False
        else:
            table_game.one_won = False
        board.close()
        return {'end': True}



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.md')) as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.txt')) as f:
    CHANGES = f.read()

requires = [
    'requests',
]

dev_require = [
    'cornice',
    'plaster_pastedeploy',
    'psycopg2-binary',
    'pyramid >= 1.9a',
    'pyramid_debugtoolbar',
    'pyramid_jinja2',
    'pyramid_retry',
    'pyramid_tm',
    'pytest',
    'pytest-cov',
    'requests',
    'SQLAlchemy',
    'transaction',
    'zope.sqlalchemy',
    'waitress',
]

tests_require = [
    'WebTest >= 1.3.1',  # py3 compat
    'pytest',
    'pytest-cov',
    'requests',
]

setup(
    name='neuralknight',
    version='1.0.1',
    description='A Chess-playing AI',
    long_description=README + '\n\n' + CHANGES,
    author='David Snowberger, Shannon Tully, and Adam Grandquist',
    author_email='david.snowberger@fartherout.org',
    url='https://www.github.com/dsnowb/neuralknight',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Games/Entertainment :: Board Games',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'License :: Freely Distributable',
        'Framework :: Pyramid',
        'Environment :: Console',
    ],
    keywords='chess entertainment game ai',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    extras_require={
        'dev': dev_require,
        'testing': tests_require,
    },
    python_requires='>=3.6',
    install_requires=requires,
    entry_points={
        'paste.app_factory': [
            'main = neuralknight:main',
        ],
        'console_scripts': [
            'initialize_neuralknight_db = neuralknight.scripts.initializedb:main',
            'neuralknight = neuralknight.scripts.cli:main',
        ],
    },
)



from .meta import Base
from sqlalchemy import (
    Column,
    Integer,
    Table,
    ForeignKey,
)


table = Table(
    'association', Base.metadata,
    Column('game_link', Integer, ForeignKey('game.id')),
    Column('board_link', Integer, ForeignKey('board.id')))



from ..models.base_agent import BaseAgent


class MockAgent(BaseAgent):
    def __init__(self, testapp, moves, game_id, player):
        self.testapp = testapp
        self.args = []
        self.kwargs = []
        self.moves = iter(moves)
        self.past_end = False
        super().__init__(game_id, player)

    def play_round(self, *args, **kwargs):
        self.args.append(args)
        self.kwargs.append(kwargs)
        try:
            return self.put_board(next(self.moves))
        except StopIteration:
            self.past_end = True
        return {}


def test_home_endpoint(testapp):
    response = testapp.get('/')
    assert response.status_code == 200


def test_games_endpoint(testapp):
    response = testapp.get('/v1.0/games')
    assert response.status_code == 200
    assert 'ids' in response.json


def test_agent_play_no_moves(testapp):
    game = testapp.post_json('/v1.0/games').json
    player1 = MockAgent(testapp, [], game['id'], 1)
    player2 = MockAgent(testapp, [], game['id'], 2)
    assert player1.agent_id != player2.agent_id
    assert player1.args.pop() == ()
    assert player1.kwargs.pop() == {}
    assert player1.past_end
    assert not player2.args
    assert not player2.kwargs
    assert not player2.past_end


# def test_agent_play_through(testapp):
#     player1_moves = [tuple(map(bytes, (
#         (12, 6, 2, 10, 4, 2, 6, 12),
#         (8, 8, 8, 8, 8, 8, 8, 8),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 0, 0, 9, 0, 0, 0),
#         (9, 9, 9, 9, 0, 9, 9, 9),
#         (13, 7, 3, 11, 5, 3, 7, 13)))), tuple(map(bytes, (
#
#         (12, 6, 2, 10, 4, 2, 6, 12),
#         (8, 8, 8, 8, 8, 0, 8, 8),
#         (0, 0, 0, 0, 0, 8, 0, 0),
#         (0, 0, 0, 0, 0, 0, 0, 11),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 0, 0, 9, 0, 0, 0),
#         (9, 9, 9, 9, 0, 9, 9, 9),
#         (13, 7, 3, 0, 5, 3, 7, 13))))]
#     player1_moves = [player1_moves[0]]
#     player2_moves = [tuple(map(bytes, (
#         (12, 6, 2, 4, 10, 2, 6, 12),
#         (8, 8, 8, 0, 8, 8, 8, 8),
#         (0, 0, 0, 8, 0, 0, 0, 0),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 9, 0, 0, 0, 0, 0),
#         (9, 9, 0, 9, 9, 9, 9, 9),
#         (13, 7, 3, 5, 11, 3, 7, 13)))), tuple(map(bytes, (
#
#         (12, 6, 2, 4, 0, 2, 6, 12),
#         (8, 8, 8, 0, 8, 8, 8, 8),
#         (0, 0, 0, 8, 0, 0, 0, 0),
#         (0, 0, 0, 0, 0, 0, 0, 0),
#         (10, 0, 0, 0, 0, 0, 0, 0),
#         (0, 0, 9, 0, 0, 0, 0, 0),
#         (9, 9, 0, 9, 9, 9, 9, 9),
#         (13, 7, 3, 5, 11, 3, 7, 13))))]
#     player2_moves = []
#     game = testapp.post_json('/v1.0/games').json
#     player1 = MockAgent(testapp, player1_moves, game['id'], 1)
#     player2 = MockAgent(testapp, player2_moves, game['id'], 2)
#     assert len(player1.args) == 1
#     assert len(player2.args) == 1
#     assert not player1.past_end
#     assert player2.past_end



from .meta import Base
from sqlalchemy.orm import relationship
from .table_ass import table
from sqlalchemy import (
    Boolean,
    Column,
    String,
    Integer,
)


class TableGame(Base):
    __tablename__ = 'game'
    id = Column(Integer, primary_key=True)
    game = Column(String, nullable=False, unique=True)
    player_one = Column(String, nullable=False, unique=True)
    player_two = Column(String, nullable=False, unique=True)
    one_won = Column(Boolean, nullable=False)
    two_won = Column(Boolean, nullable=False)
    board_link = relationship("TableBoard", secondary=table, back_populates="game_link")



import requests
import sys

from cmd import Cmd
from time import sleep

PIECE_NAME = {
    3: 'bishop',
    5: 'king',
    7: 'knight',
    9: 'pawn',
    11: 'queen',
    13: 'rook',
}
PROMPT = '> '
BRIGHT_GREEN = '\u001b[42;1m'
RESET = '\u001b[0m'
SELECTED_PIECE = f'{ BRIGHT_GREEN }{{}}{ RESET }'
TOP_BOARD_OUTPUT_SHELL = '''
  A B C D E F G H
 +---------------'''
BOARD_OUTPUT_SHELL = ('8|', '7|', '6|', '5|', '4|', '3|', '2|', '1|')


def get_info(api_url, game_id):
    response = requests.get(f'{ api_url }/v1.0/games/{ game_id }/info')
    return response.json()['print']


def format_board(board):
    return map(' '.join, board.splitlines())


def print_board(board):
    """
    Print board in shell.
    """
    print(TOP_BOARD_OUTPUT_SHELL)
    for shell, line in zip(
            BOARD_OUTPUT_SHELL, tuple(board)):
        print(f'{ shell }{ "".join(line) }')


class CLIAgent(Cmd):
    prompt = PROMPT

    def __init__(self, api_url):
        """
        Init player board.
        """
        super().__init__()
        self.api_url = api_url
        self.do_reset()

    def do_reset(self, *args):
        self.piece = None
        game = requests.post(f'{ self.api_url }/v1.0/games').json()
        game['user'] = 1
        self.game_id = game['id']
        self.user = requests.post(
            f'{ self.api_url }/issue-agent',
            json=game,
            headers={
                'content-type': 'application/json'
            },
        ).json()['agent_id']
        requests.post(
            f'{ self.api_url }/issue-agent-lookahead',
            json={'id': self.game_id, 'player': 2, 'lookahead': 4})
        print('> piece <col> <row>  # select piece')
        print('> move <col> <row>   # move selected piece to')
        print('> reset              # start a new game')
        print_board(format_board(get_info(self.api_url, self.game_id)))

    def do_piece(self, arg_str):
        """
        Select piece for move.
        """
        args = self.parse(arg_str)
        if len(args) != 2:
            return self.print_invalid('piece ' + arg_str)
        self.piece = args
        response = requests.get(f'{ self.api_url }/v1.0/games/{ self.game_id }')
        state = response.json()['state']
        if state == {'end': True}:
            return print('game over')
        board = tuple(map(bytes.fromhex, state))
        try:
            piece = board[args[1]][args[0]]
        except IndexError:
            return self.print_invalid('piece ' + arg_str)
        if not (piece and (piece & 1)):
            return self.print_invalid('piece ' + arg_str)
        board = list(map(list, get_info(self.api_url, self.game_id).splitlines()))
        board[args[1]][args[0]] = SELECTED_PIECE.format(
            board[args[1]][args[0]])
        print_board(map(' '.join, board))
        print(f'Selected: { PIECE_NAME[piece & 0xf] }')

    def do_move(self, arg_str):
        """
        Make move.
        """
        if not self.piece:
            return self.print_invalid('move ' + arg_str)

        args = self.parse(arg_str)
        if len(args) != 2:
            return self.print_invalid('move ' + arg_str)

        move = {'move': (tuple(reversed(self.piece)), tuple(reversed(args)))}
        self.piece = None

        response = requests.put(
            f'{ self.api_url }/agent/{ self.user }',
            json=move,
            headers={
                'content-type': 'application/json',
            }
        )
        if response.status_code != 200 or response.json().get('invalid', False):
            print_board(format_board(get_info(self.api_url, self.game_id)))
            return print('Invalid move.')
        if response.json().get('state', {}).get('end', False):
            print_board(format_board(get_info(self.api_url, self.game_id)))
            return print('you won')
        response = requests.get(f'{ self.api_url }/v1.0/games/{ self.game_id }')
        in_board = response.json()['state']
        print_board(format_board(get_info(self.api_url, self.game_id)))
        if in_board == {'end': True}:
            return print('you won')
        print('making move ...')
        board = in_board
        while in_board == board:
            sleep(2)
            response = requests.get(f'{ self.api_url }/v1.0/games/{ self.game_id }')
            state = response.json()['state']
            if state == {'end': True}:
                return print('game over')
            response = requests.get(
                f'{ self.api_url }/agent/{ self.user }',
                headers={
                    'content-type': 'application/json',
                }
            )
            if response.status_code != 200:
                return self.do_reset()
            try:
                if response.json()['state'] == {'end': True}:
                    return self.do_reset()
            except Exception:
                return self.do_reset()
            board = state
        print_board(format_board(get_info(self.api_url, self.game_id)))

    def print_invalid(self, args):
        print_board(format_board(get_info(self.api_url, self.game_id)))
        print('invalid command:', args)
        print('> piece <col> <row>  # select piece')
        print('> move <col> <row>   # move selected piece to')
        print('> reset              # start a new game')

    @staticmethod
    def parse(args):
        """
        Split arguments.
        """
        args = args.split()
        if len(args) != 2:
            return args
        try:
            args[1] = 8 - int(args[1])
            if not (0 <= args[1] < 8):
                print('out of range row')
                raise ValueError
        except ValueError:
            print('not int', args[1])
            return ()
        try:
            args[0] = ord(args[0]) - ord('a')
            if not (0 <= args[1] < 8):
                print('out of range column')
                raise ValueError
        except ValueError:
            print('not char', args[0])
            return ()
        return args

    def emptyline(self):
        """
        Do nothing on empty command.
        """

    def precmd(self, line):
        """
        Sanitize data.
        """
        return line.strip().lower()


def main(argv=sys.argv):
    try:
        port = 8080
        api_url = f'http://localhost:{ port }'
        if len(argv) > 1:
            api_url = argv[1]
        while True:
            try:
                CLIAgent(api_url).cmdloop()
            except Exception:
                pass
    except KeyboardInterrupt:
        print()



from cornice import Service
from operator import methodcaller
from pyramid.httpexceptions import HTTPBadRequest

from ..models.board import Board, NoBoard
from ..models.board_model import InvalidMove

games = Service(
    name='games',
    path='/v1.0/games',
    description='Create game')
game_states = Service(
    name='game_states',
    path='/v1.0/games/{game}/states',
    description='Game states')
game_interaction = Service(
    name='game_interaction',
    path='/v1.0/games/{game}',
    description='Game interaction')
game_info = Service(
    name='game_info',
    path='/v1.0/games/{game}/info',
    description='Game info')


class BlankBoard:
    def __str__(self):
        return '\n' * 8

    def add_player_v1(self, *args, **kwargs):
        return {}

    def close(self, *args, **kwargs):
        return {'end': True}

    def current_state_v1(self, *args, **kwargs):
        return {'state': {'end': True}}

    def slice_cursor_v1(self, *args, **kwargs):
        return {'cursor': None, 'boards': []}

    def update_state_v1(self, *args, **kwargs):
        return {'end': True}


def get_game(request):
    """
    Retrieve board for request.
    """
    try:
        return Board.get_game(request.matchdict['game'])
    except NoBoard:
        pass
    return BlankBoard()


@games.get()
def get_games(request):
    """
    Retrieve all game ids.
    """
    return {'ids': tuple(Board.GAMES.keys())}


@games.post()
def post_game(request):
    """
    Create a new game and provide an id for interacting.
    """
    return {'id': Board().id}


@game_states.get()
def get_states(request):
    """
    Start or continue a cursor of next board states.
    """
    cursor = get_game(request).slice_cursor_v1(**request.GET)
    cursor['boards'] = tuple(map(
        lambda boards: tuple(map(
            lambda board: tuple(map(methodcaller('hex'), board)),
            boards)),
        cursor['boards']))
    return cursor


@game_interaction.get()
def get_state(request):
    """
    Provide current state on the board.
    """
    state = get_game(request).current_state_v1()
    if isinstance(state['state'], dict):
        return state
    state['state'] = tuple(map(methodcaller('hex'), state['state']))
    return state


@game_interaction.post()
def join_game(request):
    """
    Add player to board.
    """
    try:
        user_id = request.json['id']
    except KeyError:
        raise HTTPBadRequest
    return get_game(request).add_player_v1(request.dbsession, user_id)


@game_interaction.put()
def put_state(request):
    """
    Make a move to a new state on the board.
    """
    try:
        return get_game(request).update_state_v1(request.dbsession, **request.json)
    except InvalidMove:
        pass
    return {'invalid': True}


@game_info.get()
def get_info(request):
    """
    Provide current state on the board.
    """
    return {'print': str(get_game(request))}



import os
import requests

from uuid import uuid4

from .board_model import BoardModel, CursorDelegate
import neuralknight


class NoBoard(Exception):
    pass


class BaseBoard:
    GAMES = {}
    if os.environ.get('PORT', ''):
        PORT = os.environ['PORT']
    else:
        PORT = 8080
    API_URL = 'http://localhost:{}'.format(PORT)

    @classmethod
    def get_game(cls, _id):
        """
        Provide game matching id.
        """
        if _id in cls.GAMES:
            return cls.GAMES[_id]
        raise NoBoard

    def __init__(self, board, _id=None, active_player=True):
        if _id:
            self.id = _id
        else:
            self.id = str(uuid4())
        self.GAMES[self.id] = self
        if isinstance(board, BoardModel):
            self._board = board
        else:
            self._board = BoardModel(board)
        self.board = self._board.board
        self.cursor_delegate = CursorDelegate()
        self._active_player = active_player
        self.player1 = None
        self.player2 = None

    def __bool__(self):
        """
        Ensure active player king on board.
        """
        return bool(self._board)

    def __contains__(self, piece):
        """
        Ensure piece on board.
        """
        return piece in self._board

    def __iter__(self):
        """
        Provide next boards at one lookahead.
        """
        return self._board.lookahead_boards(1)

    def request(self, method, resource, *args, json=None, **kwargs):
        if neuralknight.testapp:
            if method == 'POST':
                return neuralknight.testapp.post_json(resource, json).json
            if method == 'PUT':
                return neuralknight.testapp.put_json(resource, json).json
            if method == 'GET':
                return neuralknight.testapp.get(resource, json).json
        if method == 'POST':
            self.executor.submit(
                requests.post, f'{ self.API_URL }{ resource }', data=json, **kwargs
            ).add_done_callback(self.handle_future)
        if method == 'PUT':
            self.executor.submit(
                requests.put, f'{ self.API_URL }{ resource }', json=json, **kwargs
            ).add_done_callback(self.handle_future)
        if method == 'GET':
            self.executor.submit(
                requests.get, f'{ self.API_URL }{ resource }', data=json, **kwargs
            ).add_done_callback(self.handle_future)

    def active_player(self):
        """
        UUID of active player.
        """
        if self._active_player:
            return self.player1
        return self.player2

    def close(self):
        self.GAMES.pop(self.id, None)
        return {}

    def current_state_v1(self):
        """
        Provide REST view of game state.
        """
        return {'state': self.board}

    def handle_future(self, future):
        """
        Handle a future from and async request.
        """
        future.result().json()

    def prune_lookahead_boards(self, n=4):
        return self._board.prune_lookahead_boards(n)

    def lookahead_boards(self, n=4):
        return self._board.lookahead_boards(n)

    def poke_player(self, end, active_player=None):
        """
        Inform active player of game state.
        """
        self.request('PUT', f'/agent/{ active_player or self.active_player() }', json={'end': end})

    def update(self, state):
        """
        Validate and return new board state.
        """
        board = type(self)(
            self._board.update(tuple(map(bytes.fromhex, state))), self.id, not self._active_player)
        board.player1 = self.player1
        board.player2 = self.player2
        return board
